@@ -18,7 +18,9 @@ from torch import nn
 
 import l2hmc.configs as cfgs
 from l2hmc.group.u1.pytorch.group import U1Phase
+from l2hmc.lattice.u1.pytorch.lattice import LatticeU1
 from l2hmc.group.su3.pytorch.group import SU3
+from l2hmc.lattice.su3.pytorch.lattice import LatticeSU3
 from l2hmc.network.pytorch.network import NetworkFactory
 
 log = logging.getLogger(__name__)
@@ -106,12 +108,23 @@ def grab(x: Tensor) -> Array:
     return x.detach().cpu().numpy()
 
 
+def dummy_network(inputs: tuple[Tensor, Tensor]):
+    x, v = inputs
+    # assert x.shape == v.shape
+    return (
+        torch.zeros_like(v),
+        torch.zeros_like(v),
+        torch.zeros_like(v),
+    )
+
+
 class Dynamics(nn.Module):
     def __init__(
             self,
             potential_fn: Callable,
             config: cfgs.DynamicsConfig,
-            network_factory: NetworkFactory
+            # backend: str = 'horovod',
+            network_factory: Optional[NetworkFactory] = None,
     ):
         """Initialization method."""
         super().__init__()
@@ -124,30 +137,35 @@ class Dynamics(nn.Module):
 
         self.network_factory = network_factory
         self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
-        self.networks = self._build_networks(network_factory)
-        self.xnet = self.networks['xnet']
-        self.vnet = self.networks['vnet']
-        self.register_module('xnet', self.networks['xnet'])
-        self.register_module('vnet', self.networks['vnet'])
-        # assert isinstance(self.networks, nn.ModuleDict)
-        # self.xnet = self.networks.get_submodule['xnet']
-        # self.vnet = self.networks.get_submodule['vnet']
-        # num_nets = self.nlf if self.config.use_separate_networks else 1
-        # self.networks = network_factory.build_networks(
-        #     n=num_nets, split_xnets=self.config.use_split_xnets
-        # )
+        if network_factory is not None:
+            self._networks_built = True
+            self.networks = self._build_networks(network_factory)
+            self.xnet = self.networks['xnet']
+            self.vnet = self.networks['vnet']
+            self.register_module('xnet', self.networks['xnet'])
+            self.register_module('vnet', self.networks['vnet'])
+            log.debug('Built networks.')
+        else:
+            self._networks_built = False
+            self.xnet = dummy_network
+            self.vnet = dummy_network
+            self.networks = {
+                'xnet': self.xnet,
+                'vnet': self.vnet,
+            }
+
+        log.debug(f'dynamics._networks_built: {self._networks_built}')
         self.masks = self._build_masks()
 
         if self.config.group == 'U1':
             self.g = U1Phase()
+            self.lattice = LatticeU1(self.config.nchains,
+                                     self.config.latvolume)
         elif self.config.group == 'SU3':
             self.g = SU3()
+            self.lattice = LatticeSU3(self.config.nchains,
+                                      self.config.latvolume)
 
-        # self.xeps = nn.ParameterDict()
-        # self.veps = nn.ParameterDict()
-        # rg = (not self.config.eps_fixed)
-        # xeps = {}
-        # veps = {}
         self.xeps = nn.ParameterList([
             nn.parameter.Parameter(
                 torch.tensor(
@@ -171,8 +189,9 @@ class Dynamics(nn.Module):
         if torch.cuda.is_available():
             self.xeps = self.xeps.cuda()
             self.veps = self.veps.cuda()
-            self.xnet.cuda()
-            self.vnet.cuda()
+            if network_factory is not None and self._networks_built:
+                self.xnet.cuda()
+                self.vnet.cuda()
             # self.networks.cuda()
             self.cuda()
 
@@ -181,7 +200,6 @@ class Dynamics(nn.Module):
         split = self.config.use_split_xnets
         n = self.nlf if self.config.use_separate_networks else 1
         networks = network_factory.build_networks(n, split)
-        # return networks['xnet'], networks['vnet']
         return networks
 
     @torch.no_grad()
@@ -322,7 +340,7 @@ class Dynamics(nn.Module):
 
     def forward(
             self,
-            inputs: tuple[Tensor, Tensor]  # x, beta
+            inputs: tuple[Tensor, Tensor]
     ) -> tuple[Tensor, dict]:
         if self.config.merge_directions:
             outputs = self.apply_transition_fb(inputs)
@@ -336,7 +354,7 @@ class Dynamics(nn.Module):
 
     def apply_transition_hmc(
             self,
-            inputs: tuple[Tensor, Tensor],  # x, beta
+            inputs: tuple[Tensor, Tensor],
             eps: Optional[float] = None,
             nleapfrog: Optional[int] = None,
     ) -> tuple[Tensor, dict]:
@@ -344,8 +362,6 @@ class Dynamics(nn.Module):
         ma_, mr_ = self._get_accept_masks(data['metrics']['acc'])
         ma_ = ma_.to(inputs[0].device)
         mr_ = mr_.to(inputs[0].device)
-        # ma = ma_.unsqueeze(-1)
-        # mr = mr_.unsqueeze(-1)
         ma = ma_[:, None]
         mr = mr_[:, None]
 
@@ -356,8 +372,6 @@ class Dynamics(nn.Module):
 
         vout = ma * vprop + mr * vinit
         xout = ma * xprop + mr * xinit
-        # if isinstance(self.g, g.SU3):
-        #     xout = self.g.vec_to_group(xout)
 
         state_out = State(x=xout, v=vout, beta=data['init'].beta)
         mc_states = MonteCarloStates(init=data['init'],
@@ -372,7 +386,7 @@ class Dynamics(nn.Module):
 
     def apply_transition_fb(
             self,
-            inputs: tuple[Tensor, Tensor]  # x, beta
+            inputs: tuple[Tensor, Tensor]
     ) -> tuple[Tensor, dict]:
         data = self.generate_proposal_fb(inputs)
         ma_, mr_ = self._get_accept_masks(data['metrics']['acc'])
@@ -383,8 +397,14 @@ class Dynamics(nn.Module):
 
         # NOTE: We construct output states by combining
         #   output = (acc_mask * proposed) + (reject_mask * init)
-        v_out = ma * data['proposed'].v + mr * data['init'].v
-        x_out = ma * data['proposed'].x + mr * data['init'].x
+        v_out = (
+            ma * data['proposed'].v.flatten(1)
+            + mr * data['init'].v.flatten(1)
+        )
+        x_out = (
+            ma * data['proposed'].x.flatten(1)
+            + mr * data['init'].x.flatten(1)
+        )
 
         # NOTE: sumlogdet = (accept * logdet) + (reject * 0)
         sumlogdet = ma_ * data['metrics']['sumlogdet']
@@ -394,6 +414,7 @@ class Dynamics(nn.Module):
                                      proposed=data['proposed'],
                                      out=state_out)
         data['metrics'].update({
+            'beta': data['init'].beta,
             'acc_mask': ma_,
             'sumlogdet': sumlogdet,
             'mc_states': mc_states,
@@ -416,8 +437,6 @@ class Dynamics(nn.Module):
         if torch.cuda.is_available():
             ma_ = ma_.cuda()
             mr_ = mr_.cuda()
-        # ma = ma_.unsqueeze(-1)
-        # mr = mr_.unsqueeze(-1)
         ma = ma_[:, None]
         mr = mr_[:, None]
 
@@ -431,20 +450,18 @@ class Dynamics(nn.Module):
                                      out=state_out)
 
         metrics.update({
+            'beta': beta,
             'acc': metrics['acc'],
             'acc_mask': ma_,
             'sumlogdet': logdet,
             'mc_states': mc_states,
         })
 
-        # metrics.update(**{f'fwd/{key}': val for key, val in mfwd.items()})
-        # metrics.update(**{f'bwd/{key}': val for key, val in mbwd.items()})
-
         return x_out, metrics
 
     def apply_transition_both(
             self,
-            inputs: tuple[Tensor, Tensor]  # x, beta
+            inputs: tuple[Tensor, Tensor]
     ) -> tuple[Tensor, dict]:
         x, beta = inputs
         fwd = self.generate_proposal(inputs, forward=True)
@@ -454,8 +471,6 @@ class Dynamics(nn.Module):
         if torch.cuda.is_available():
             mf_ = mf_.cuda()
             mb_ = mb_.cuda()
-        # mf_ = mf_.to(x.device)
-        # mb_ = mb_.to(x.device)
         mf = mf_[:, None]
         mb = mb_[:, None]
 
@@ -483,11 +498,8 @@ class Dynamics(nn.Module):
         if torch.cuda.is_available():
             ma_ = ma_.cuda()
             mr_ = mr_.cuda()
-        # ma = ma_.unsqueeze(-1)
-        # mr = mr_.unsqueeze(-1)
         ma = ma_[:, None]
         mr = mr_[:, None]
-        # mr = mr_.unsqueeze(-1)
 
         v_out = ma * vp + mr * v_init
         x_out = ma * xp + mr * x
@@ -515,9 +527,6 @@ class Dynamics(nn.Module):
             'mc_states': mc_states,
         })
 
-        # metrics.update(**{f'fwd/{key}': val for key, val in mfwd.items()})
-        # metrics.update(**{f'bwd/{key}': val for key, val in mbwd.items()})
-
         return x_out, metrics
 
     def random_state(self, beta: float) -> State:
@@ -535,15 +544,13 @@ class Dynamics(nn.Module):
 
     def generate_proposal_hmc(
             self,
-            inputs: tuple[Tensor, Tensor],  # x, beta
+            inputs: tuple[Tensor, Tensor],
             eps: Optional[float] = None,
             nleapfrog: Optional[int] = None,
     ) -> dict:
         x, beta = inputs
-        # v = torch.randn_like(x).to(x.device)
         xshape = [x.shape[0], *self.xshape[1:]]
         v = self.g.random_momentum(xshape).to(x.device)
-        # v = v.reshape_as(x).to(x.device)
 
         init = State(x=x, v=v, beta=beta)
         proposed, metrics = self.transition_kernel_hmc(
@@ -556,13 +563,11 @@ class Dynamics(nn.Module):
 
     def generate_proposal_fb(
             self,
-            inputs: tuple[Tensor, Tensor],  # x, beta
+            inputs: tuple[Tensor, Tensor],
     ) -> dict:
         x, beta = inputs
         xshape = [x.shape[0], *self.xshape[1:]]
         v = self.g.random_momentum(xshape).to(x.device)
-        # v = self.g.random_momentum(list(x.shape)).to(x.device)
-        # v = torch.randn_like(x).to(x.device)
         init = State(x=x, v=v, beta=beta)
         proposed, metrics = self.transition_kernel_fb(init)
 
@@ -570,14 +575,12 @@ class Dynamics(nn.Module):
 
     def generate_proposal(
             self,
-            inputs: tuple[Tensor, Tensor],  # x, beta
+            inputs: tuple[Tensor, Tensor],
             forward: bool,
     ) -> dict:
         x, beta = inputs
         xshape = [x.shape[0], *self.xshape[1:]]
         v = self.g.random_momentum(xshape).to(x.device)
-        # v = self.g.random_momentum(list(self.xshape)).to(x.device)
-        # v = torch.randn_like(x).to(x.device)
         state_init = State(x=x, v=v, beta=beta)
         state_prop, metrics = self.transition_kernel(state_init, forward)
 
@@ -638,6 +641,56 @@ class Dynamics(nn.Module):
         v2 = v1 - 0.5 * eps * force2                       # v -= ½ veps * f
         return State(x=xp, v=v2, beta=state.beta)          # output: (x', v')
 
+    def transition_kernel_hmc1(
+            self,
+            state: State,
+            eps: Optional[float] = None,
+            nleapfrog: Optional[int] = None,
+    ) -> tuple[State, dict]:
+        state_ = State(x=state.x, v=state.v, beta=state.beta)
+        sumlogdet = torch.zeros(state.x.shape[0],
+                                # dtype=state.x.real.dtype,
+                                device=state.x.device)
+        history = {}
+        if self.config.verbose:
+            history = self.update_history(
+                self.get_metrics(state_, sumlogdet),
+                history=history,
+            )
+        eps = self.config.eps_hmc if eps is None else eps
+        nlf = (
+            self.config.nleapfrog if not self.config.merge_directions
+            else 2 * self.config.nleapfrog
+        )
+        assert nlf <= 2 * self.config.nleapfrog
+        nleapfrog = nlf if nleapfrog is None else nleapfrog
+
+        x = self.g.update_gauge(state.x, 0.5 * eps * state.v)
+        force = self.grad_potential(x, state.beta)
+        v = state.v - eps * force
+        for _ in range(1, nleapfrog):
+            x = self.g.update_gauge(x, eps * v)
+            force = self.grad_potential(x, state.beta)
+            v = v - eps * force
+            state_ = State(x=x, v=v, beta=state.beta)
+            if self.config.verbose:
+                history = self.update_history(
+                    self.get_metrics(state_, sumlogdet),
+                    history=history
+                )
+
+        x = self.g.update_gauge(x, 0.5 * eps * v)
+        state_ = State(x=x, v=v, beta=state.beta)
+        acc = self.compute_accept_prob(state, state_, sumlogdet)
+        history.update({'acc': acc, 'sumlogdet': sumlogdet})
+
+        if self.config.verbose:
+            for key, val in history.items():
+                if isinstance(val, list) and isinstance(val[0], Tensor):
+                    history[key] = torch.stack(val)
+
+        return state_, history
+
     def transition_kernel_hmc(
             self,
             state: State,
@@ -646,7 +699,7 @@ class Dynamics(nn.Module):
     ) -> tuple[State, dict]:
         state_ = State(x=state.x, v=state.v, beta=state.beta)
         sumlogdet = torch.zeros(state.x.shape[0],
-                                dtype=state.x.real.dtype,
+                                # dtype=state.x.real.dtype,
                                 device=state.x.device)
         history = {}
         if self.config.verbose:
@@ -684,9 +737,11 @@ class Dynamics(nn.Module):
             self,
             state: State,
     ) -> tuple[State, dict]:
-        sumlogdet = torch.zeros((state.x.shape[0],),
-                                # dtype=state.x.real.dtype,
-                                device=state.x.device)
+        sumlogdet = torch.zeros(
+            (state.x.shape[0],),
+            dtype=state.x.dtype,
+            device=state.x.device,
+        )
         sldf = torch.zeros_like(sumlogdet)
         sldb = torch.zeros_like(sumlogdet)
 
@@ -718,7 +773,6 @@ class Dynamics(nn.Module):
         # Flip momentum
         m1 = -1.0 * torch.ones_like(state_.v)
         state_ = State(state_.x, (m1 * state_.v), state_.beta)
-        # sumlogdet *= -1
 
         # Backward
         for step in range(self.config.nleapfrog):
@@ -791,19 +845,15 @@ class Dynamics(nn.Module):
     ) -> Tensor:
         h_init = self.hamiltonian(state_init)
         h_prop = self.hamiltonian(state_prop)
+        if sumlogdet.is_complex():
+            log.warning('Complex sumlogdet! Taking norm...?')
+            sumlogdet = sumlogdet.norm()
+
         dh = h_init - h_prop + sumlogdet
         prob = torch.exp(
             torch.minimum(dh, torch.zeros_like(dh, device=dh.device))
         ).to(state_init.x.device)
 
-        # return prob
-        # return torch.where(
-        #     torch.isfinite(prob),
-        #     prob,
-        #     torch.zeros_like(prob)
-        # )
-        # return torch.where(prob.nan_to_num)
-        # return prob.nan_to_num(0.)
         return prob
 
     @staticmethod
@@ -812,13 +862,6 @@ class Dynamics(nn.Module):
             px > torch.rand_like(px).to(px.device)
         ).to(torch.float)
         rej = torch.ones_like(acc) - acc
-
-        # runif = torch.rand_like(px)  # .to(torch.float)
-        # acc = (px > runif).to(torch.float)
-        # rej = torch.ones_like(acc) - acc
-        # acc = (px > torch.rand_like(px).to(px.device)).to(torch.float)
-        # rej = torch.ones_like(acc) - acc
-        # return acc.to(px.device), rej.to(px.device)
         return acc, rej
 
     @staticmethod
@@ -843,22 +886,28 @@ class Dynamics(nn.Module):
             mask = np.zeros((self.xdim,), dtype=np.float32)
             mask[idx] = 1.
             masks.append(torch.from_numpy(mask[None, :]))
-            # masks.append(torch.from_numpy(mask[None, :]).to(self.device))
 
-        # return torch.stack(list(masks)).to(self.device)
-        # return torch.from_numpy(np.array(masks)).float().to(self.device)
         return masks
 
-    def _get_vnet(self, step: int) -> nn.Module:
+    def _get_vnet(self, step: int) -> nn.Module | Callable:
         """Returns momentum network to be used for updating v."""
+        if not self._networks_built:
+            return self.vnet
         # vnet = self.networks.get_submodule('vnet')
         if self.config.use_separate_networks:
             return self.vnet.get_submodule(str(step))
         return self.vnet
 
-    def _get_xnet(self, step: int, first: bool = False) -> nn.Module:
+    def _get_xnet(
+            self,
+            step: int,
+            first: bool = False
+    ) -> nn.Module | Callable:
         """Returns position network to be used for updating x."""
         # xnet = self.networks.get_submodule('xnet')
+        if not self._networks_built:
+            return self.xnet
+
         if self.config.use_separate_networks:
             xnet = self.xnet.get_submodule(str(step))
             if self.config.use_split_xnets:
@@ -871,15 +920,12 @@ class Dynamics(nn.Module):
     def _stack_as_xy(self, x: Tensor) -> Tensor:
         """Returns -pi < x <= pi stacked as [cos(x), sin(x)]"""
         # TODO: Deal with ConvNet here
-        # if self.config.use_conv_net:
-        #     xcos = xcos.reshape(self.lattice_shape)
-        #     xsin = xsin.reshape(self.lattice_shape)
         return torch.stack([x.cos(), x.sin()], dim=-1).to(self.device)
 
     def _call_vnet(
             self,
             step: int,
-            inputs: tuple[Tensor, Tensor],  # (x, ∂S/∂x)
+            inputs: tuple[Tensor, Tensor],
     ) -> tuple[Tensor, Tensor, Tensor]:
         """Call the momentum update network used to update v.
 
@@ -888,24 +934,18 @@ class Dynamics(nn.Module):
         Returns:
             s, t, q: Scaling, Translation, and Transformation functions
         """
-        vnet = self._get_vnet(step)
-        assert callable(vnet)
         x, force = inputs
-        # if isinstance(self.g, U1Phase):
-        # x = self._stack_as_xy(x)
-        # elif isinstance(self.g, SU3):
-        # x = self.g.group_to_vec(x)
-
-        # x, force = x.to(self.device), force.to(self.device)
         if torch.cuda.is_available():
             x, force = x.cuda(), force.cuda()
+        vnet = self._get_vnet(step)
+        assert callable(vnet)
 
         return vnet((x, force))
 
     def _call_vnet_dummy(
             self,
             step: int,
-            inputs: tuple[Tensor, Tensor],  # (x, ∂S/∂x)
+            inputs: tuple[Tensor, Tensor],
     ) -> tuple[Tensor, Tensor, Tensor]:
         """Call the momentum update network used to update v.
 
@@ -944,7 +984,6 @@ class Dynamics(nn.Module):
         x, v = inputs
         x = self.g.group_to_vec(x)
 
-        # x, v = x.to(self.device), v.to(self.device)
         if torch.cuda.is_available():
             x, v = x.cuda(), v.cuda()
 
@@ -958,9 +997,10 @@ class Dynamics(nn.Module):
             self,
             step: int,
             inputs: tuple[Tensor, Tensor],
-            first: bool = False,
+            first: bool = False
     ) -> tuple[Tensor, Tensor, Tensor]:
-        """Call the position network used to update x.
+        """
+        Call the position network used to update x.
 
         Args:
             inputs: (m * x, v) tuple, where (m * x) is a masking operation.
@@ -970,12 +1010,15 @@ class Dynamics(nn.Module):
         xnet = self._get_xnet(step, first)
         assert callable(xnet)
         x, v = inputs
-        x = self.g.group_to_vec(x)
-
-        # x, v = x.to(self.device), v.to(self.device)
         if torch.cuda.is_available():
             x, v = x.cuda(), v.cuda()
 
+        if isinstance(self.g, SU3):
+            x = x.reshape(self.xshape)
+            x = torch.stack([x.real, x.imag], 1)
+        elif isinstance(self.g, U1Phase):
+            x = self.g.group_to_vec(x.reshape(-1, *self.xshape[1:]))
+
         return xnet((x, v))
 
     def _forward_lf(self, step: int, state: State) -> tuple[State, Tensor]:
@@ -983,7 +1026,6 @@ class Dynamics(nn.Module):
         m, mb = self._get_mask(step)
         m, mb = m.to(self.device), mb.to(self.device)
         sumlogdet = torch.zeros(state.x.shape[0],
-                                # dtype=state.x.real.dtype,
                                 device=self.device)
 
         state, logdet = self._update_v_fwd(step, state)
@@ -1008,7 +1050,6 @@ class Dynamics(nn.Module):
         m, mb = self._get_mask(step_r)
         m, mb = m.to(self.device), mb.to(self.device)
         sumlogdet = torch.zeros((state.x.shape[0],),
-                                # dtype=state.x.real.dtype,
                                 device=self.device)
 
         state, logdet = self._update_v_bwd(step_r, state)
@@ -1030,14 +1071,14 @@ class Dynamics(nn.Module):
         assert isinstance(self.veps, nn.ParameterList)
         eps = self.veps[step]
         force = self.grad_potential(state.x, state.beta)
+        force = force.reshape_as(state.v)
         s, t, q = self._call_vnet(step, (state.x, force))
 
-        jac = eps * s / 2.  # jacobian factor, also used in exp_s below
-        logdet = jac.sum(dim=1)
-        exp_s = jac.exp()
-        exp_q = (eps * q).exp()
-        # exp_s = torch.exp(jac)
-        # exp_q = torch.exp(eps * q)
+        logjac = eps * s / 2.  # jacobian factor, also used in exp_s below
+        logdet = logjac.flatten(1).sum(dim=1)
+        exp_s = logjac.exp().reshape_as(state.v)
+        exp_q = (eps * q).exp().reshape_as(force)
+        t = t.reshape_as(force)
         vf = exp_s * state.v - 0.5 * eps * (force * exp_q + t)
 
         return State(state.x, vf, state.beta), logdet
@@ -1050,9 +1091,10 @@ class Dynamics(nn.Module):
         s, t, q = self._call_vnet(step, (state.x, force))
 
         logjac = (-eps * s / 2.)  # jacobian factor, also used in exp_s below
-        logdet = logjac.sum(dim=1)
-        exp_s = torch.exp(logjac)
-        exp_q = torch.exp(eps * q)
+        logdet = logjac.flatten(1).sum(dim=1)
+        exp_s = torch.exp(logjac).reshape_as(state.v)
+        exp_q = torch.exp(eps * q).reshape_as(force)
+        t = t.reshape_as(force)
         vb = exp_s * (state.v + 0.5 * eps * (force * exp_q + t))
 
         return State(state.x, vb, state.beta), logdet
@@ -1068,38 +1110,42 @@ class Dynamics(nn.Module):
         assert isinstance(self.xeps, nn.ParameterList)
         eps = self.xeps[step]
         mb = (torch.ones_like(m) - m).to(self.device)
-        xm_init = m * state.x
+        x = state.x.flatten(1)
+        v = state.v.reshape_as(x)
+        xm_init = m * x
         inputs = (xm_init, state.v)
         s, t, q = self._call_xnet(step, inputs, first=first)
-        # s, t, q = self._call_xnet_dummy(step, inputs, first=first)
         s = eps * s
         q = eps * q
         exp_s = s.exp()
         exp_q = q.exp()
         if isinstance(self.g, U1Phase):
             if self.config.use_ncp:
-                halfx = state.x / 2.
+                halfx = (x / 2.).flatten(1)
                 _x = 2. * (halfx.tan() * exp_s).atan()
-                xp = _x + eps * (state.v * exp_q + t)
+                xp = _x + eps * (v * exp_q + t)
                 xf = xm_init + (mb * xp)
                 cterm = (halfx.cos()) ** 2
                 sterm = (exp_s * halfx.sin()) ** 2
                 logdet_ = (exp_s / (cterm + sterm)).log()
                 logdet = (mb * logdet_).sum(dim=1)
             else:
-                xp = state.x * exp_s + eps * (state.v * exp_q + t)
+                xp = x * exp_s + eps * (v * exp_q + t)
                 xf = xm_init + (mb * xp)
                 logdet = (mb * s).sum(dim=1)
 
         elif isinstance(self.g, SU3):
-            x = self.g.group_to_vec(state.x)
+            x = state.x.reshape(self.xshape)
+            xm_init = xm_init.reshape(self.xshape)
+            exp_s = exp_s.reshape(self.xshape)
+            exp_q = exp_q.reshape(self.xshape)
+            t = t.reshape(self.xshape)
+            # x = self.g.group_to_vec(state.x)
             # v = self.g.group_to_vec(state.v)
-            xm_init = self.g.group_to_vec(xm_init)
+            # xm_init = self.g.group_to_vec(xm_init)
             xp = x * exp_s + eps * (state.v * exp_q + t)
-            xf = xm_init + (mb * xp)
-            # xp = state.x * exp_s + eps * (state.v * exp_q + t)
-            # xf = xm_init + (mb * xp)
-            logdet = (mb * s).sum(dim=1)
+            xf = xm_init + (mb * xp.flatten(1)).reshape_as(xm_init)
+            logdet = (mb * s.flatten(1)).sum(dim=1).real
         else:
             raise ValueError('Unexpected value for `self.g`')
 
@@ -1117,25 +1163,25 @@ class Dynamics(nn.Module):
         assert isinstance(self.xeps, nn.ParameterList)
         eps = self.xeps[step]
         mb = (torch.ones_like(m) - m).to(self.device)
-        xm_init = m * state.x
-        inputs = (xm_init, state.v)
+        x = state.x.flatten(1)
+        v = state.v.reshape_as(x)
+        xm_init = m * x
+        inputs = (xm_init, v)
+        # xm_init = (m * state.x.flatten(1)).reshape_as(state.x)
+        # inputs = (xm_init, state.v)
         s, t, q = self._call_xnet(step, inputs, first=first)
-        # s, t, q = self._call_xnet_dummy(step, inputs, first=first)
         s = (-eps) * s
         q = eps * q
         exp_s = s.exp()
         exp_q = q.exp()
-        # exp_s = torch.exp(s)
-        # exp_q = torch.exp(q)
         if isinstance(self.g, U1Phase):
             if self.config.use_ncp:
-                halfx = state.x / 2.
+                halfx = x / 2.
                 halfx_scale = exp_s * halfx.tan()
                 x1 = 2. * halfx_scale.atan()
-                x2 = exp_s * eps * (state.v * exp_q + t)
+                x2 = exp_s * eps * (v * exp_q + t)
                 xnew = x1 - x2
                 xb = xm_init + (mb * xnew)
-
                 cterm = halfx.cos() ** 2
                 sterm = (exp_s * halfx.sin()) ** 2
                 logdet_ = (exp_s / (cterm + sterm)).log()
@@ -1145,13 +1191,26 @@ class Dynamics(nn.Module):
                 xb = xm_init + (mb * xnew)
                 logdet = (mb * s).sum(dim=1)
         elif isinstance(self.g, SU3):
-            xnew = exp_s * (state.x - eps * (state.v * exp_q + t))
-            xb = xm_init + (mb * xnew)
-            logdet = (mb * s).sum(dim=1)
+            t = t.reshape_as(state.x).to(state.x.dtype)
+            exp_s = exp_s.reshape_as(state.x).to(state.x.dtype)
+            exp_q = exp_q.reshape_as(state.x).to(state.x.dtype)
+            eps = eps.to(state.x.dtype)
+            xnew = exp_s * self.g.update_gauge(
+                state.x,
+                -(eps * (state.v * exp_q + t))
+            )
+            xb = (xm_init + (mb * xnew.flatten(1))).reshape(self.xshape)
+            logdet = (mb * s.flatten(1).to(mb.dtype)).sum(1).real
+
+            # xnew = exp_s * (state.x - eps * (state.v * exp_q + t))
+            # xmb = (mb * xnew.flatten(1)).reshape_as(state.x)
+            # xb = xm_init.reshape_as(state.x) + xmb  # .reshape_as(xm_init)
+            # logdet = (mb * s).sum(dim=1)
+            # import pdb; pdb.set_trace()
         else:
             raise ValueError('Unexpected value for `self.g`')
 
-        # xb = self.g.compat_proj(xb)
+        xb = self.g.compat_proj(xb)
         return State(x=xb, v=state.v, beta=state.beta), logdet
 
     def hamiltonian(self, state: State) -> Tensor:
@@ -1162,29 +1221,16 @@ class Dynamics(nn.Module):
 
     def kinetic_energy(self, v: Tensor) -> Tensor:
         """Returns the kinetic energy, KE = 0.5 * v ** 2."""
-        # return 0.5 * v.reshape(v.shape[0], -1).square()
         return self.g.kinetic_energy(v)
 
     def potential_energy(self, x: Tensor, beta: Tensor):
         """Returns the potential energy, PE = beta * action(x)."""
-        # return beta * self.potential_fn(x)
         return self.potential_fn(x, beta)
 
     def grad_potential(
             self,
             x: Tensor,
             beta: Tensor,
-            # create_graph: bool = True,
     ) -> Tensor:
         """Compute the gradient of the potential function."""
-        x = x.to(self.device)
-        x.requires_grad_(True)
-        s = self.potential_energy(x, beta)
-        id = torch.ones(x.shape[0], device=self.device)
-        dsdx, = torch.autograd.grad(s, x,
-                                    create_graph=True,
-                                    retain_graph=True,
-                                    # create_graph=create_graph,
-                                    # retain_graph=True,
-                                    grad_outputs=id)
-        return dsdx
+        return self.lattice.grad_action(x, beta)
