@@ -7,32 +7,23 @@ Author: Sam Foreman
 Date: 06/02/2021
 """
 from __future__ import absolute_import, annotations, division, print_function
-from dataclasses import dataclass
+# from dataclasses import dataclass
 from typing import Optional
 from math import pi as PI
 
 from scipy.special import i0, i1
 import torch
 
-from l2hmc.lattice.u1.numpy.lattice import BaseLatticeU1
+from l2hmc.lattice.lattice import Lattice
+# from l2hmc.lattice.u1.numpy.lattice import BaseLatticeU1
+import l2hmc.group.u1.pytorch.group as g
+# import l2hmc.group.pytorch.group as g
+from l2hmc.configs import Charges, LatticeMetrics
 
-TWO_PI = 2. * PI
+TWOPI = 2. * PI
 Tensor = torch.Tensor
 
 
-@dataclass
-class Charges:
-    intQ: Tensor
-    sinQ: Tensor
-
-
-@dataclass
-class LatticeMetrics:
-    plaqs: Tensor
-    charges: Charges
-    p4x4: Tensor
-
-
 def area_law(beta: float, nplaqs: int):
     return (i1(beta) / i0(beta)) ** nplaqs
 
@@ -43,25 +34,60 @@ def plaq_exact(beta: float | Tensor):
 
 def project_angle(x: Tensor) -> Tensor:
     """For x in [-4pi, 4pi], returns x in [-pi, pi]."""
-    return x - TWO_PI * torch.floor((x + PI) / TWO_PI)
+    return x - TWOPI * torch.floor((x + PI) / TWOPI)
 
 
-class LatticeU1(BaseLatticeU1):
-    def __init__(self, nb: int, shape: tuple[int, int]):
-        super().__init__(nb, shape=shape)
+class LatticeU1(Lattice):
+    def __init__(self, nchains: int, shape: list[int]):
+        assert len(shape) == 2
+        self.g = g.U1Phase()
+        self.nt, self.nx, = shape
+        self.volume = self.nt * self.nx
+        self.nplaqs = self.nt * self.nx
+        # self.nsites = np.cumprod(shape)[-1]
+        # self.nlinks = self.nsites * self.g._dim
+        # self.site_idxs = tuple(
+        #     [self.nt]
+        #     + [self.nx for _ in range(self.g._dim - 1)]
+        # )
+        # self.link_idxs = tuple(list(self.site_idxs) + [self.g._dim])
+
+        super().__init__(group=self.g, nchains=nchains, shape=shape)
 
     def draw_uniform_batch(self, requires_grad=True) -> Tensor:
-        """Draw batch of samples, uniformly from [-pi, pi)."""
-        unif = torch.rand(self._shape, requires_grad=requires_grad)
-        return TWO_PI * unif - PI
+        """Draw batch of samples, uniformly from (-pi, pi)."""
+        return TWOPI * (
+            torch.rand(self._shape, requires_grad=requires_grad)
+        ) - PI
 
-    def unnormalized_log_prob(self, x: Tensor, beta: Tensor) -> Tensor:
-        return self.action(x=x, beta=beta)
+    def kinetic_energy(self, v: Tensor) -> Tensor:
+        return 0.5 * v.flatten(1) ** 2
 
     def action(self, x: Tensor, beta: Tensor) -> Tensor:
-        """Calculate the Wilson gauge action for a batch of lattices."""
         wloops = self._get_wloops(x)
-        return beta * (1. - torch.cos(wloops)).sum((1, 2))
+        return self._action(wloops, beta)
+
+    def _action(
+            self,
+            wloops: Tensor,
+            beta: Tensor
+    ) -> Tensor:
+        """Calculate the Wilson gauge action for a batch of lattices."""
+        return beta * (1. - wloops.cos()).sum((1, 2))
+
+    def action_with_grad(
+            self,
+            x: Tensor,
+            beta: Tensor
+    ) -> tuple[Tensor, Tensor]:
+        x.requires_grad_(True)
+        s = self.action(x, beta)
+        identity = torch.ones(x.shape[0], device=x.device)
+        assert isinstance(s, Tensor)
+        dsdx, = torch.autograd.grad(s, x,
+                                    retain_graph=True,
+                                    grad_outputs=identity)
+        return s, dsdx
 
     def grad_action(
             self,
@@ -73,9 +99,10 @@ class LatticeU1(BaseLatticeU1):
         x.requires_grad_(True)
         s = self.action(x, beta)
         identity = torch.ones(x.shape[0], device=x.device)
+        assert isinstance(s, Tensor)
         dsdx, = torch.autograd.grad(s, x,
-                                    retain_graph=True,
                                     create_graph=create_graph,
+                                    retain_graph=True,
                                     grad_outputs=identity)
         return dsdx
 
@@ -94,73 +121,58 @@ class LatticeU1(BaseLatticeU1):
     def calc_metrics(
             self,
             x: Tensor,
-            beta: Optional[float] = None
     ) -> dict[str, Tensor]:
         """Calculate various metrics and return as dict"""
         wloops = self.wilson_loops(x)
         plaqs = self.plaqs(wloops=wloops)
         charges = self.charges(wloops=wloops)
-        metrics = {'plaqs': plaqs}
-        if beta is not None:
-            metrics.update({
-               'plaqs_err': plaq_exact(torch.from_numpy(beta)) - plaqs
-            })
-        metrics.update({
-            'intQ': charges.intQ, 'sinQ': charges.sinQ
-        })
-
-        return metrics
+        return {
+            'plaqs': plaqs,
+            'intQ': charges.intQ,
+            'sinQ': charges.sinQ
+        }
 
     def observables(self, x: Tensor) -> LatticeMetrics:
         """Calculate observables and return as LatticeMetrics object"""
         wloops = self.wilson_loops(x)
-        return LatticeMetrics(p4x4=self.plaqs4x4(x=x),
-                              plaqs=self.plaqs(wloops=wloops),
-                              charges=self.charges(wloops=wloops))
+        return LatticeMetrics(
+            p4x4=self.plaqs4x4(x=x),
+            plaqs=self.plaqs(wloops=wloops),
+            charges=self.charges(wloops=wloops)
+        )
 
     def wilson_loops(self, x: Tensor) -> Tensor:
         """Calculate the Wilson loops by summing links in CCW direction."""
-        # --------------------------
-        # NOTE: Watch your shapes!
-        # --------------------------
-        # * First, x.shape = [-1, 2, Lt, Lx], so
-        #       (x_reshaped).T.shape = [2, Lx, Lt, -1]
-        #   and,
-        #       x0.shape = x1.shape = [Lx, Lt, -1]
-        #   where x0 and x1 are the links along the 2 (t, x) dimensions.
-        #
-        # * The Wilson loop is then:
-        #       wloop = U0(x, y) +  U1(x+1, y) - U0(x, y+1) - U(1)(x, y)
-        #   and so output = wloop.T, with output.shape = [-1, Lt, Lx]
-        # --------------------------
-        x0, x1 = x.reshape(-1, *self.xshape).transpose(0, 1)
-        x0 = x0.T
-        x1 = x1.T
-
-        return (x0 + x1.roll(-1, dims=0) - x0.roll(-1, dims=1) - x1).T
+        x = x.view((-1, *self.xshape))  # shape: [nchains, 2, Nt, Nx]
+        xu = x[:, 0]  # shape: [nchains, Nt, Nx]
+        xv = x[:, 1]  # shape: [nchains, Nt, Nx]
+        return xu + xv.roll(-1, dims=1) - xu.roll(-1, dims=2) - xv
 
     def wilson_loops4x4(self, x: Tensor) -> Tensor:
         """Calculate the 4x4 Wilson loops"""
-        x0, x1 = x.reshape(-1, *self.xshape).transpose(0, 1)
-        x0 = x0.T
-        x1 = x1.T
+        x = x.reshape(-1, *self.xshape)
+        xu = x[:, 0]
+        xv = x[:, 1]
+        # xu, xv = x.reshape(-1, *self.xshape).transpose(0, 1)
+        # xv = xv.T
+        # xu = xu.T
         return (
-            x0                                  # Ux  [x, y]
-            + x0.roll(-1, dims=2)               # Ux  [x+1, y]
-            + x0.roll(-2, dims=2)               # Ux  [x+2, y]
-            + x0.roll(-3, dims=2)               # Ux  [x+3, y]
-            + x0.roll(-4, dims=2)               # Ux  [x+4, y]
-            + x1.roll((-4, -1), dims=(2, 1))    # Uy  [x+4, y+1]
-            + x1.roll((-4, -2), dims=(2, 1))    # Uy  [x+4, y+2]
-            + x1.roll((-4, -3), dims=(2, 1))    # Uy  [x+4, y+3]
-            - x0.roll((-3, -4), dims=(2, 1))    # -Ux [x+3, y+4]
-            - x0.roll((-2, -4), dims=(2, 1))    # -Ux [x+2, y+4]
-            - x0.roll((-1, -4), dims=(2, 1))    # -Ux [x+1, y+4]
-            - x1.roll(-4, dims=1)               # -Uy [x, y+4]
-            - x1.roll(-3, dims=1)               # -Uy [x, y+3]
-            - x1.roll(-2, dims=1)               # -Uy [x, y+2]
-            - x1.roll(-1, dims=1)               # -Uy [x, y+1]
-            - x1                                # -Uy [x, y]
+            xu                                  # Ux  [x, y]
+            + xu.roll(-1, dims=2)               # Ux  [x+1, y]
+            + xu.roll(-2, dims=2)               # Ux  [x+2, y]
+            + xu.roll(-3, dims=2)               # Ux  [x+3, y]
+            + xu.roll(-4, dims=2)               # Ux  [x+4, y]
+            + xv.roll((-4, -1), dims=(2, 1))    # Uy  [x+4, y+1]
+            + xv.roll((-4, -2), dims=(2, 1))    # Uy  [x+4, y+2]
+            + xv.roll((-4, -3), dims=(2, 1))    # Uy  [x+4, y+3]
+            - xu.roll((-3, -4), dims=(2, 1))    # -Ux [x+3, y+4]
+            - xu.roll((-2, -4), dims=(2, 1))    # -Ux [x+2, y+4]
+            - xu.roll((-1, -4), dims=(2, 1))    # -Ux [x+1, y+4]
+            - xv.roll(-4, dims=1)               # -Uy [x, y+4]
+            - xv.roll(-3, dims=1)               # -Uy [x, y+3]
+            - xv.roll(-2, dims=1)               # -Uy [x, y+2]
+            - xv.roll(-1, dims=1)               # -Uy [x, y+1]
+            - xv                                # -Uy [x, y]
         ).T
 
     def plaqs(
@@ -175,10 +187,13 @@ class LatticeU1(BaseLatticeU1):
                 raise ValueError('One of `x` or `wloops` must be specified.')
             wloops = self.wilson_loops(x)
 
-        return torch.cos(wloops).mean((1, 2))
+        return wloops.cos().mean((1, 2))
+
+    def _plaqs(self, wloops: Tensor) -> Tensor:
+        return wloops.cos().mean((1, 2))
 
     def _plaqs4x4(self, wloops4x4: Tensor) -> Tensor:
-        return torch.cos(wloops4x4).mean((1, 2))
+        return wloops4x4.cos().mean((1, 2))
 
     def plaqs4x4(
             self,
@@ -195,11 +210,12 @@ class LatticeU1(BaseLatticeU1):
 
     def _sin_charges(self, wloops: Tensor) -> Tensor:
         """Calculate sinQ from Wilson loops."""
-        return torch.sin(wloops).sum((1, 2)) / TWO_PI
+        # return torch.sin(wloops).sum((1, 2)) / TWO_PI
+        return wloops.sin().sum((1, 2)) / TWOPI
 
     def _int_charges(self, wloops: Tensor) -> Tensor:
         """Calculate intQ from Wilson loops."""
-        return project_angle(wloops).sum((1, 2)) / TWO_PI
+        return (project_angle(wloops)).sum((1, 2)) / TWOPI
 
     def _get_wloops(
             self,
@@ -209,6 +225,12 @@ class LatticeU1(BaseLatticeU1):
             raise ValueError('One of `x` or `wloops` must be specified.')
         return self.wilson_loops(x)
 
+    def random(self) -> Tensor:
+        return self.g.random(list(self._shape))
+
+    def random_momentum(self) -> Tensor:
+        return self.g.random_momentum(list(self._shape))
+
     def sin_charges(
             self,
             x: Optional[Tensor] = None,
@@ -238,6 +260,11 @@ class LatticeU1(BaseLatticeU1):
         intQ = self._int_charges(wloops)
         return Charges(intQ=intQ, sinQ=sinQ)
 
+    def _charges(self, wloops: Tensor) -> Charges:
+        sinQ = self._sin_charges(wloops)
+        intQ = self._int_charges(wloops)
+        return Charges(intQ=intQ, sinQ=sinQ)
+
     def plaq_loss(
             self,
             acc: Tensor,
@@ -245,11 +272,12 @@ class LatticeU1(BaseLatticeU1):
             x2: Optional[Tensor] = None,
             wl1: Optional[Tensor] = None,
             wl2: Optional[Tensor] = None,
-    ) -> float:
+    ) -> Tensor:
         wloops1 = self._get_wloops(x1) if wl1 is None else wl1
         wloops2 = self._get_wloops(x2) if wl2 is None else wl2
-        dwloops = 2. * (1. - torch.cos(wloops2 - wloops1))
+        dwloops = 2. * (1. - (wloops2 - wloops1).cos())
         ploss = acc * dwloops.sum((1, 2)) + 1e-4
+        assert isinstance(ploss, Tensor)
 
         return -ploss.mean(0)
 
@@ -260,7 +288,7 @@ class LatticeU1(BaseLatticeU1):
             x2: Optional[Tensor] = None,
             wl1: Optional[Tensor] = None,
             wl2: Optional[Tensor] = None,
-    ):
+    ) -> Tensor:
         wloops1 = self._get_wloops(x1) if wl1 is None else wl1
         wloops2 = self._get_wloops(x2) if wl2 is None else wl2
         q1 = self._sin_charges(wloops=wloops1)
@@ -268,3 +296,12 @@ class LatticeU1(BaseLatticeU1):
         dq = (q2 - q1) ** 2
         qloss = acc * dq + 1e-4
         return -qloss.mean(0)
+
+
+if __name__ == '__main__':
+    lattice = LatticeU1(3, [8, 8])
+    beta = torch.tensor(1.0)
+    x = lattice.random()
+    v = lattice.random_momentum()
+    action = lattice.action(x, beta)
+    kinetic = lattice.kinetic_energy(v)
